/*
 * CSR BlueCore ROM version initialisation
 * Copyright by TomTom BV
 * Written by Christian Daniel 2004 with some inspiration from
 * OpenBT - copyright by Axis Communication AB
 * pskey - copyright by Fabrizio Gennari
 * published under GPLv2 or newer
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ubcsp.h"
#include "bt_usb.h"

#undef  DEBUG
#undef  RFBOARD_RESET

#ifdef DEBUG
#define DBG(fmt, arg...) do { printf(fmt ,##arg); fflush(stdout); } while (0)
#else
#define DBG(fmt, arg...) do {} while (0)
#endif

enum BCCMD_MODE {
	BCCMDMODE_H4,
	BCCMDMODE_UBCSP,
	BCCMDMODE_USB
};

int mode;
int crystal_freq = 16000000;
int ext_clock = 0; /* set of no ana_ftrim possible */
int detected_chiprevision;
int csr_build_id;

/* sending any powertable seems to crash CSR ROM in some cases */
#define SEND_POWERTABLE

/******************************************************************************/

/* file descriptor for serial port */
int serialfd;

/*
 * open serial port with following parameters: xxxxxx baud, 8E1, no flow control
 */
int serial_open(const char *devname, speed_t speed)
{
	int fd;
	struct termios ti;
	unsigned long arg;

	if((fd = open(devname, O_RDWR | O_NOCTTY)) < 0) {
		warn("Could not open serial port %s", devname);
		return -1;
	}
	tcflush(fd, TCIOFLUSH);

	if(tcgetattr(fd, &ti) < 0) {
		warn("Could not get serial port attributes");
		goto failed;
	}

	/* be sure that serial port is configured correctly */
	cfmakeraw(&ti);
	ti.c_cflag |=  CLOCAL;
	ti.c_cflag &= ~CRTSCTS;
	ti.c_cflag |=  PARENB;
	ti.c_cflag &= ~PARODD;
	ti.c_cflag &= ~CSIZE;
	ti.c_cflag |=  CS8;
	ti.c_cflag &= ~CSTOPB;

	/* one byte at a time, no timer */
	ti.c_cc[VMIN] = 1;
	ti.c_cc[VTIME] = 0;

	cfsetspeed(&ti, speed);

	if(tcsetattr(fd, TCSANOW, &ti) < 0) {
		warn("Could net set serial port attributes");
		goto failed;
	}

	/* non blocking i/o is needed for the state machines below */
	arg = fcntl(fd, F_GETFL, 0);
	arg |= O_NONBLOCK;
	if(fcntl(fd, F_SETFL, arg) < 0) {
		warn("Could not set serial port to nonblocking");
		goto failed;
	}

	return fd;

failed:
	close(fd);
	return -1;
}

/*
 * send a single byte to the serial port - being called externally from ubcsp
 */
void put_uart(uint8 ch)
{
	write(serialfd, &ch, 1);
}

/*
 * get a byte from the serial port - being called externally from ubcsp
 */
uint8 get_uart(uint8 *ch)
{
	int res;

	res = read(serialfd, ch, 1);
	return res > 0 ? res : 0;
}

/******************************************************************************/

#ifdef RFBOARD_RESET

#define GPS_DRIVER_MAGIC	'U'
#define IOW_GPS_ON		_IO(GPS_DRIVER_MAGIC, 1)
#define IOW_GPS_OFF		_IO(GPS_DRIVER_MAGIC, 2)
#define IOW_GPS_RESET		_IO(GPS_DRIVER_MAGIC, 3)

/*
 * resets the RF board by cycling the power and asserting the nReset line
 */
void reset_rf_board()
{
	int fd;

	DBG("resetting RF board\n");
	fd = open("/dev/gps", O_RDWR);
	ioctl(fd, IOW_GPS_OFF, NULL);
	sleep(1);
	ioctl(fd, IOW_GPS_ON, NULL);
	ioctl(fd, IOW_GPS_RESET, NULL);
	sleep(1);
	close(fd);
}

#endif /* RFBOARD_RESET */

/******************************************************************************/

/* HCI command packet header */
struct cmd_pkt {
	uint16 opcode;
	uint8 len;
	uint8 data[256];
} __attribute__((packed));

/* CSR proprietary message header */
struct csr_msg {
	uint8 desc;
	uint8 data[0];
} __attribute__((packed));

/* CSR variable set- or get-header */
struct csr_bccmd {
	uint16 type;
	uint16 len;
	uint16 seq;
	uint16 id;
	uint16 status;
	uint16 data[0];
} __attribute__((packed));

/* CSR ps key access header */
struct csr_bccmd_ps {
	uint16 ps_key;
	uint16 ps_len;
	uint16 unused;
	uint16 ps_val[0];
} __attribute__((packed));

/* loads of magic defines */
#define CMD_HDR_LEN		3
#define HCI_HDR_LEN		1
#define BCSP_CMD_CHN		5
#define MANUFACTURER_SPEC	0x3f
#define CSR_CH_ID_BCCMD		0x02

#define CSR_MSGTYPE_GETREQ	0x0000
#define CSR_MSGTYPE_SETREQ	0x0002
#define CSR_CMD_BUILD_ID	0x2819
#define CSR_CMD_CHIPVER		0x281a
#define CSR_CMD_CHIPREV		0x281b
#define CSR_CMD_PS_SIZE		0x3006
#define CSR_CMD_PS		0x7003
#define CSR_WARM_RESET		0x4002

#define CSR_STATUS_OK		0x0000

#define CSR_PS_ANA_FREQ			0x01fe
#define CSR_PS_ANA_FTRIM		0x01f6
#define CSR_PS_TX_OFFSET_HALF_MHZ	0x0217
#define CSR_PS_CLASSOFDEVICE		0x0003
#define CSR_PS_FEATURES			0x00EF
#define CSR_PS_COUNTRYCODE		0x0002
#define CSR_PS_HOST_INTERFACE		0x01f9
#define CSR_PS_LC_DEFAULT_TX_POWER	0x0021
#define CSR_PS_UART_BAUD_RATE		0x01be
#define CSR_PS_UART_CONFIG_FLOW_CTRL_EN	0x03fe
#define CSR_PS_AMUX_B			0x022c
#define CSR_PS_PATCH_22 		0x00ac
#define CSR_PS_TXRX_PIO_CONTROL		0x0209
#define CSR_PS_BDADDR			0x0001
#define CSR_PS_BCCMD_SECURITY_ACTIVE	0x01fd
#define CSR_PS_LC_POWER_TABLE		0x001e
#define CSR_PS_HOSTIO_MAP_SCO_PCM	0x01ab
#define CSR_PS_USB_VENDORID		0x02be
#define CSR_PS_USB_PRODUCTID		0x02bf
#define	CSR_PS_HOSTIO_SCO_HCI_THRESHOLDS 0x01bc
#define CSR_PS_MAX_SCOS 		0x000e


int real_baudrate;
uint16 csr_div;
int csr_rate;

#define hci_put_opcode(ocf, ogf) (cpu_to_le16(((ocf) & 0x03ff) | \
	((ogf) << 10 & 0xfc00)))

#define GET_BITS(var, bit, bits) (((var) >> (bit)) & ((1UL << (bits)) - 1))
#define SET_BITS(var, bit, bits, x)					\
	do {								\
		(var) = ((var) & ~(((1UL << (bits)) - 1) << (bit))) |	\
		(((x) & ((1UL << (bits)) - 1)) << (bit));		\
	} while (0)

#define CSR_GET_LAST(msg)	GET_BITS((msg)->desc, 7, 1)
#define CSR_SET_LAST(msg, x)	SET_BITS((msg)->desc, 7, 1, (x) != 0)
#define CSR_SET_FIRST(msg, x)	SET_BITS((msg)->desc, 6, 1, (x) != 0)
#define CSR_GET_CH_ID(msg)	GET_BITS((msg)->desc, 0, 6)
#define CSR_SET_CH_ID(msg, id)	SET_BITS((msg)->desc, 0, 6, id)

#if (__BYTE_ORDER == __LITTLE_ENDIAN)
#	undef __BIG_ENDIAN
#endif
#if (__BYTE_ORDER == __BIG_ENDIAN)
#	undef __LITTLE_ENDIAN
#endif

#ifdef __LITTLE_ENDIAN
#	define cpu_to_le16(x) (x)
#	define cpu_to_le32(x) (x)
#	define cpu_to_be16(x) htons((x))
#	define cpu_to_be32(x) htonl((x))
#else
#	define cpu_to_be16(x) (x)
#	define cpu_to_be32(x) (x)
	extern inline __u16 cpu_to_le16(__u16 x) { return (x<<8) | (x>>8); }
	extern inline __u32 cpu_to_le32(__u32 x) { return ((x>>24) |
		((x>>8)&0xff00) | ((x<<8)&0xff0000) | (x<<24)); }
#endif

#define le16_to_cpu(x)  cpu_to_le16(x)
#define le32_to_cpu(x)  cpu_to_le32(x)
#define be16_to_cpu(x)  cpu_to_be16(x)
#define be32_to_cpu(x)  cpu_to_be32(x)

char *csr_chipver[] = {
	"BlueCore01a",
	"BlueCore01b",
	"BlueCore2-External, BlueCore2-ROM (ST or TSMC)",
	"BlueCore3-MM, -ROM, -Flash or BlueCore4-External, -ROM",
	"BlueCore unknown"
};

struct csr_chiprev {
	uint16 value;
	char *desc;
};

#define CHIPREVISION_BLUECORE3_ROM 0x15
#define CHIPREVISION_BLUECORE4_ROM 0x30

struct csr_chiprev csr_chiprev[] = {
	{0x64, "BlueCore01b, engineering sample (ES) version"},
	{0x65, "BlueCore01b, production version"},
	{0x89, "BlueCore2-External, version a"},
	{0x8a, "BlueCore2-External, version b"},
	{0x28, "BlueCore2-ROM"},
	{0x43, "BlueCore3-MM"},
	{CHIPREVISION_BLUECORE3_ROM, "BlueCore3-ROM"},
	{0xe2, "BlueCore3-Flash"},
	{0x26, "BlueCore4-External"},
	{CHIPREVISION_BLUECORE4_ROM, "BlueCore4-ROM"},
	{0x00, "BlueCore unknown"}
};


struct power_entry {
	uint8 external_pa;
	uint8 internal_pa;
	uint8 reserved;
	int8 tx_dbm;
} __attribute__((__packed__));

#define MAX_PT_ENTRIES 16

struct btcal_flash {
	uint16 crystal_trim;
	union {
		uint32 raw;
		struct power_entry u;
	} power_table[MAX_PT_ENTRIES + 1];
} __attribute__((__packed__));

struct btcal_flash btcal_flash;

int btcal_sanity_check()
{
	int i;
	int last_dbm;
	int err;
	const int min_step_size = 2;
	const int max_step_size = 8;
	if (btcal_flash.power_table[0].raw == 0x00000000) {
		printf("sanity check: powertable[0] is 0x00000000, must be 0xffffffff to indicate no powertable\n");
		return -1;
	}
	if (btcal_flash.power_table[0].u.tx_dbm >= 4) {
		printf("sanity check: first entry powertable[0].tx_dbm %d (0x%02x) must be < 4 dBm\n",
			btcal_flash.power_table[0].u.tx_dbm, (uint8)btcal_flash.power_table[0].u.tx_dbm);
		return -1;
	}
	last_dbm = 0; // will not be used, 
	for (i=0; i<(MAX_PT_ENTRIES+1); i++) {
		err = 0;
		/* check if entry is end of list */
		if ((i > 0) && (btcal_flash.power_table[i].raw == 0x00000000)) {
			printf("sanity check: proper 0x00000000 end of table at [%d]\n",i);
			if (i == 1) {
				printf("sanity check: powertable is only one entry long, BAD\n");
				return -1;
			}
			return 0;
		}
		if (btcal_flash.power_table[i].raw == 0xffffffff) {
			printf("sanity check: raw 0xffffffff powertable[%d] entry found, table must be 0x00000000 terminated\n",i);
			return -1;
		}
		if (btcal_flash.power_table[i].u.internal_pa > 0x3f) {
			printf("sanity check: powertable[%d].internal_pa 0x%02x must be below 0x3f\n",i,btcal_flash.power_table[i].u.internal_pa);
			err = -1;
		}
		if (btcal_flash.power_table[i].u.external_pa != 0x00) {
			printf("sanity check: powertable[%d].external_pa 0x%02x must be 0x00, cant support external amp\n",i,btcal_flash.power_table[i].u.external_pa);
			err = -1;
		}
		if (btcal_flash.power_table[i].u.reserved != 0x00) {
			printf("sanity check: powertable[%d].reserved 0x%02x must be 0x00, invalid entry\n",i,btcal_flash.power_table[i].u.reserved);
			err = -1;
		}
		if (i > 0) {
			// since last_dbm is initially not set
			if (btcal_flash.power_table[i].u.tx_dbm <= (last_dbm + min_step_size)) {
				printf("sanity check: powertable[%d] == %d (%02x), must be at least 2 dBm higher as previous entry\n",i,btcal_flash.power_table[i].u.tx_dbm,(uint8)btcal_flash.power_table[i].u.tx_dbm);
				err = -1;
			}
			if (btcal_flash.power_table[i].u.tx_dbm >= (last_dbm + max_step_size)) {
				printf("sanity check: powertable[%d] == %d (%02x), can be max 8 dBm higher as previous entry\n",i,btcal_flash.power_table[i].u.tx_dbm,(uint8)btcal_flash.power_table[i].u.tx_dbm);
				err = -1;
			}
		}
		if (btcal_flash.power_table[i].u.tx_dbm < -30) {
			printf("sanity check: WARNING powertable[%d] %d (%02x) is below -30 dBm, out of spec\n",i,btcal_flash.power_table[i].u.tx_dbm,(uint8)btcal_flash.power_table[i].u.tx_dbm);
		}
		// update last_sdm
		last_dbm = btcal_flash.power_table[i].u.tx_dbm;
	}
	if (err == -1)
		return -1;
	printf("sanity check: too many entries (%d), table must be 0x00000000 terminated\n",i);
	return -1;
}

// =============================================================================
// 
// Created by patchmaker.pl $Revision: 1.57.2.2 $ $Date: 2007/05/24 18:24:33 $
// Creation date: Fri Mar  2 16:44:48 2007
// 
// Patch is for build 1958 odj_4hci_rom_bt2.0_19p2_0503071059_encr128
// Input file: build_1958_patches.xap.
// Depot file: //depot/bc/19p2/dev/patches/1958/build_1958_patches.xap
// File revision: 10
// 
// This patch is B-10910. It fixes B-8062, B-6989, B-8710, B-6002, B-6334,
// B-21798 and optionally B-5329 and B-4509.
// 
// B-8062
// A race condition has been identified within the BlueCore firmware which
// can cause packet transmission failure under some circumstances. This 
// causes reduced link performance.
// 
// B-6989
// Enhancements to the EDR temperature compensation algorithm to reduce the drop
// in EDR output power at high temperatures. Must be used with
// PSKEY_TEMPERATURE_VS_DELTA_TX_BB_MR_PAYLOAD &
// PSKEY_TEMPERATURE_VS_DELTA_TX_BB_MR_HEADER containing entries at 54 and 55?C.
// The recommended settings for these PSKEYS are 
//  For PSKEY_TEMPERATURE_VS_DELTA_TX_BB_MR_PAYLOAD:
//   { -30, 8 }, { 0, 4 }, { 25, 0 }, { 54, 0 }, { 55, 10 }, { 85, 4 },
//     { 105, 4 }
//  or as raw words:
//   0xffe2, 0x0008, 0x0000, 0x0004, 0x0019, 0x0000, 0x0036, 0x0000, 0x0037,
//   0x000a, 0x0055, 0x0004, 0x0069, 0x0004 
// 
//  For PSKEY_TEMPERATURE_VS_DELTA_TX_BB_MR_HEADER:
//    { -30, 5 }, { 0, 4 }, { 25, 0 }, { 54, -4 }, { 55, 5 }, { 85, 4 },
//     { 105, 4 }
//  or as raw words:
//   0xffe2, 0x0005, 0x0000, 0x0004, 0x0019, 0x0000, 0x0036, 0xfffc, 0x0037,
//   0x0005, 0x0055, 0x0004, 0x0069, 0x0004
// 
// B-8710
// Data stalls have been observed for EDR devices placed within a few
// centimetres of each other, where BlueCore cannot receive the EDR packet and
// so continuously NAKs or does not respond. This is caused by a fault in the
// chip's automatic gain control (AGC) software. The stall eventually clears,
// but has been observed to cause up to 5 seconds of delay on an EDR connection
// when data is transfered before power control has had a chance to reduce the
// sender's transmit power. The stall will only occur when the device has
// performed an inquiry since it was last reset.
// 
// This patch assumes that PSKEY_RX_ATTEN_BACKOFF and
// PSKEY_RX_ATTEN_UPDATE_RATE remain set to their default values.
// 
// B-5329
// An issue has been identified in BlueCore ROM parts which are used in a
// configuration such that the host's clock request is an input on PIO[3]
// and is combined internally with BlueCore's clock request logic; the
// resulting composite clock request being an output on PIO[2]. This is the
// behavior enabled by setting PSKEY_CLOCK_REQUEST_ENABLE to 3.
// 
// Unfortunately, the clock request output will glitch to "inactive" (0)
// for around 3uS, around 200ms after reset. This may cause issues within
// the host system, because a typical use case will be to boot the Bluecore
// device, download host-specific configuration data (such as a Bluetooth
// address) and then perform a warm reset.
// 
// This patch will restore the correct behaviour, which is that the clock
// request output from the chip will be placed into a Hi-Z state for the
// duration of the reset, to be pulled by an external resistor, and will
// not be actively driven low.
// 
// This patch is commented-out by default in this patch file, because not
// every application uses this clock request functionality.  To enable the
// patch, uncomment the patch:
// 
// 	patch patch_38, PSKEY_PATCH38 (0x00bc, 188) 
// 
// And, using the host software, also set PSKEY_CLOCK_REQUEST_ENABLE
// (0x246, 582) to zero.  In effect, the patch disables the code wired to
// PSKEY_CLOCK_REQUEST_ENABLE and replaces it with code which asserts the
// clock requst signal at a later point in the chip boot sequence.  The
// drive direction of the clock request pin is thereby set up correctly
// before the drive to the pin is enabled, thus removing the glitch.  
// 
// B-6002, B-6334
// When BlueCore is on link with EDR eSCO but not EDR ACL, EDR eSCO packets are
// not transmitted or received correctly. This is caused by an oversight in the
// firmware, which fails to enable certain hardware required for EDR. The
// supplied patch remedies this. Also, BlueCore does not disable the hardware
// when an EDR eSCO or ACL link is disconnected, which uses power unnecessarily.
// The patch also remedies this.
// 
// B-4509
// When Bluecore is configured to use the H4DS host transport, it is 
// possible for the host interface to stall. This happens following certain
// patterns of UART activity which can cause Bluecore to enter an infinite
// cycle of sending You-May-Sleep and Wake-Up messages. This patch prevents
// the problem from occurring.
// 
// NOTE: This patch, PSKEY_PATCH37 (0x00BB), is only applicable to the H4DS
// host transport. The patch need not be applied when using any other host
// transport.
// 
// B-21798
// When we receive an inbound connection from a device that we think is already
// connected we drop the old connection and use the new one in preference. This
// is because typically what has happened is the old device has been reset and
// has reconnected before the supervision timeout for the old connection.
// 
// Unfortunately, when this happens, we don't freeze the old HCI handle (see
// PSKEY_HCI_HANDLE_FREEZE_PERIOD). If the host writes to the old handle after
// we've disconnected it but, presumably, before the host has processed the
// disconnection complete event, then the chip will panic.
// 
// This patch ensures that the old handle is frozen correctly as it would be
// after a real supervision timeout.

// patch_spare5, PSKEY_PATCH34 (0x00b8, 184)
/* &00b8 = 0002 8416 031b 4216 22f4 4316 0684 1ff0 9015 08c4 16f0 9511 9515 fdc4 9525 9015 08b4 9025 9715 01b4 9725 ff00 6f15 fdc4 ff00 6f25 ff00 5315 20b4 ff00 5325 9521 0118 ff2b ff0e 8400 1218 00e2 0113 3f30 e721 0118 ff2b ff0e 8400 1918 00e2 5d46 */

// patch_spare4, PSKEY_PATCH33 (0x00b7, 183)
/* &00b7 = 0003 3041 0014 fe27 0114 ff27 0218 ff2b ff0e 3000 4418 00e2 63c2 */

// patch_spare6, PSKEY_PATCH35 (0x00b9, 185)
/* &00b9 = 0002 b6bf 0117 b800 e184 07f4 b800 f184 04f4 b800 1784 12f0 e300 b519 06f4 1412 15b2 0cf0 001a fcf0 fc3c 0014 0327 0118 ff2b ff0e 4400 7818 00e2 043c fe0e ffe3 8cd3 */

// patch_bc_boot, PSKEY_PATCH1 (0x0097, 151)
/* &0097 = 045c 032b 020a 0014 0027 2100 b434 0018 012b 010e f100 4018 009e 0084 10f4 f225 f219 0116 0012 001b 003b 003b fd00 7a38 0022 0126 f215 a000 fe54 0226 0017 0134 0027 0884 e3fc 2100 bc14 0018 012b 010e f100 4018 009e 0084 08f4 1410 0123 010e a000 0054 0127 019f 020e 043c ffe3 2a15 */

// patch_spare7, PSKEY_PATCH36 (0x00ba, 186)
/* &00ba = 0001 03f3 1414 de00 bc25 e000 6b15 a000 e354 de00 bd25 0114 fe00 8c25 0400 3e14 fe00 8d25 fe00 8e15 3134 fe00 8e25 0018 ff2b ff0e 0400 f918 00e2 ff2b 0084 11ec 9811 4000 00c0 e921 0010 9025 9815 4000 00c4 e9d5 05f0 0130 4000 0180 f8fc ffe3 0008 0000 ffe3 0114 fe00 8c25 0400 f314 fe00 8d25 fe00 8e15 3154 fe00 8e25 de00 bd19 a000 1338 0814 0026 ffe3 389c */

// patch_spare3, PSKEY_PATCH32 (0x00b6, 182)
/* &00b6 = 0001 23bf e700 b015 04f0 ff00 ffc0 03e0 0100 00b0 e721 e515 0018 ff2b ff0e 2400 d718 00e2 7d30 */

// patch_38, PSKEY_PATCH38 (0x00bc, 188)
// &00bc = 9511 fdc0 9515 9521 9711 4000 00b0 9721 6111 04b0 6121 ff00 5211 01b0 ff00 5221 9525 00e2 db18

// patch_spare2, PSKEY_PATCH31 (0x00b5, 181)
/* &00b5 = 0001 f4c6 fe0a ff2b fb3c 3784 03e4 0010 02e0 0110 e700 b021 0018 ff2b ff0e f500 c918 00e2 86b9 */

// patch_37, PSKEY_PATCH37 (0x00bb, 187)
/* &00bb = 0002 0ec1 0018 ff2b ff0e 6800 a718 009e 023c fe0e ffe3 ead4 */

// patch_radiotest_tx_start, PSKEY_PATCH17 (0x00a7, 167)
/* &00a7 = 0513 00f6 ff00 5311 e700 b015 04f0 ff00 ffc0 03e0 0100 00b0 ff00 5321 00e2 bc45 */

// patch_lc_rx_hw_set, PSKEY_PATCH25 (0x00af, 175)
/* &00af = 0915 0184 00f2 0214 0925 00e2 f019 */

// patch_spare1, PSKEY_PATCH30 (0x00b4, 180)
/* &00b4 = 0002 b264 1414 0118 ff2b ff0e b200 6718 00e2 27ff */


unsigned char patch_22[37*2] = {
0x00,0x00, 0x2b,0xff, 0x14,0x00, 0x00,0xff, 0x25,0x7b, 0x15,0x80, 0x00,0x10, 0xb4,0x00,
0x25,0x80, 0x14,0x04, 0x00,0xff, 0x25,0x79, 0x00,0x03, 0x14,0x20, 0x21,0xe3, 0x34,0xff,
0xf0,0xff, 0x00,0x45, 0x14,0xac, 0x00,0xff, 0x25,0x7b, 0x00,0x3f, 0x14,0x80, 0x34,0xff,
0xf0,0xff, 0x14,0x01, 0x00,0xff, 0x25,0x79, 0x14,0x00, 0x00,0xff, 0x25,0x7b, 0x15,0x80,
0x00,0xf0, 0xc4,0xff, 0x25,0x80, 0xe3,0xff, 0x00,0x00 
};


// patch_spare5, PSKEY_PATCH34 (0x00b8, 184)
/* &00b8 = 0002 8416 031b 4216 22f4 4316 0684 1ff0 9015 08c4 16f0 9511 9515 fdc4 9525 9015 08b4 9025 9715 01b4 9725 ff00 6f15 fdc4 ff00 6f25 ff00 5315 20b4 ff00 5325 9521 0118 ff2b ff0e 8400 1218 00e2 0113 3f30 e721 0118 ff2b ff0e 8400 1918 00e2 5d46 */

// patch_spare4, PSKEY_PATCH33 (0x00b7, 183)
/* &00b7 = 0003 3041 0014 fe27 0114 ff27 0218 ff2b ff0e 3000 4418 00e2 63c2 */

// patch_spare6, PSKEY_PATCH35 (0x00b9, 185)
/* &00b9 = 0002 b6bf 0117 b800 e184 07f4 b800 f184 04f4 b800 1784 12f0 e300 b519 06f4 1412 15b2 0cf0 001a fcf0 fc3c 0014 0327 0118 ff2b ff0e 4400 7818 00e2 043c fe0e ffe3 8cd3 */

// patch_bc_boot, PSKEY_PATCH1 (0x0097, 151)
/* &0097 = 045c 032b 020a 0014 0027 2100 b434 0018 012b 010e f100 4018 009e 0084 10f4 f225 f219 0116 0012 001b 003b 003b fd00 7a38 0022 0126 f215 a000 fe54 0226 0017 0134 0027 0884 e3fc 2100 bc14 0018 012b 010e f100 4018 009e 0084 08f4 1410 0123 010e a000 0054 0127 019f 020e 043c ffe3 2a15 */

// patch_spare7, PSKEY_PATCH36 (0x00ba, 186)
/* &00ba = 0001 03f3 1414 de00 bc25 e000 6b15 a000 e354 de00 bd25 0114 fe00 8c25 0400 3e14 fe00 8d25 fe00 8e15 3134 fe00 8e25 0018 ff2b ff0e 0400 f918 00e2 ff2b 0084 11ec 9811 4000 00c0 e921 0010 9025 9815 4000 00c4 e9d5 05f0 0130 4000 0180 f8fc ffe3 0008 0000 ffe3 0114 fe00 8c25 0400 f314 fe00 8d25 fe00 8e15 3154 fe00 8e25 de00 bd19 a000 1338 0814 0026 ffe3 389c */

// patch_spare3, PSKEY_PATCH32 (0x00b6, 182)
/* &00b6 = 0001 23bf e700 b015 04f0 ff00 ffc0 03e0 0100 00b0 e721 e515 0018 ff2b ff0e 2400 d718 00e2 7d30 */

// patch_38, PSKEY_PATCH38 (0x00bc, 188)
// &00bc = 9511 fdc0 9515 9521 9711 4000 00b0 9721 6111 04b0 6121 ff00 5211 01b0 ff00 5221 9525 00e2 db18

// patch_spare2, PSKEY_PATCH31 (0x00b5, 181)
/* &00b5 = 0001 f4c6 fe0a ff2b fb3c 3784 03e4 0010 02e0 0110 e700 b021 0018 ff2b ff0e f500 c918 00e2 86b9 */

// patch_37, PSKEY_PATCH37 (0x00bb, 187)
/* &00bb = 0002 0ec1 0018 ff2b ff0e 6800 a718 009e 023c fe0e ffe3 ead4 */

// patch_radiotest_tx_start, PSKEY_PATCH17 (0x00a7, 167)
/* &00a7 = 0513 00f6 ff00 5311 e700 b015 04f0 ff00 ffc0 03e0 0100 00b0 ff00 5321 00e2 bc45 */

// patch_lc_rx_hw_set, PSKEY_PATCH25 (0x00af, 175)
/* &00af = 0915 0184 00f2 0214 0925 00e2 f019 */

// patch_spare1, PSKEY_PATCH30 (0x00b4, 180)
/* &00b4 = 0002 b264 1414 0118 ff2b ff0e b200 6718 00e2 27ff */

/* command sequence counter */
unsigned short csr_seq;

/* sending state */
int state;

/* result of last PS key size message */
int pskey_size;

/******************************************************************************/

/*
 * process an event received from the CSR module
 */
void process_csr_event(struct csr_bccmd *cmd)
{
	int i;

#ifdef DEBUG
	printf("type:    %d\n", cmd->type);
	printf("len:     %d\n", cmd->len);
	printf("seq:     %d\n", cmd->seq);
	printf("id:      %04x\n", cmd->id);
	printf("status:  %d\n", cmd->status);
	printf("data:    ");
	for(i = 0; i < cmd->len - 5; i++)
		printf("%04x ", cmd->data[i]);
	printf("\n");
#endif /* DEBUG */
	if(cmd->status != CSR_STATUS_OK)
		printf("!!!! CSR MODULE ON FIRE %d !!!!\n", cmd->status);

	switch(cmd->id) {
		case CSR_CMD_BUILD_ID:
			csr_build_id = cmd->data[0];
			printf("CSR BuildID: %04x\n", csr_build_id);
			break;

		case CSR_CMD_CHIPVER:
			printf("CSR chip version: %s\n",
				csr_chipver[cmd->data[0] <= 3 ?
				cmd->data[0] : 4]);
			break;

		case CSR_CMD_CHIPREV:
			detected_chiprevision = cmd->data[0];
			for(i = 0; csr_chiprev[i].value != 00; i++) {
				if(csr_chiprev[i].value == cmd->data[0]) {
					printf("CSR chip revision: %s\n",
						csr_chiprev[i].desc);
					return;
				}
			}
			printf("CSR chip revision: %s\n",
				csr_chiprev[i].desc);
			break;

		case CSR_CMD_PS_SIZE:
			printf("CSR pskey size: key %04x - size %d\n",
				cmd->data[0], cmd->data[1]);
			/* save it for later use */
			pskey_size = cmd->data[1];
			break;
	}
}

/*
 * process a packet received from the module
 */
void process_packet(struct 
ubcsp_packet *rpacket)
{
#ifdef DEBUG
	int i;
#endif /* DEBUG */
	struct csr_msg *msg;
	struct csr_bccmd *cmd;

#ifdef DEBUG
	printf("\n--- received packet: ---\n");
	printf("channel:  %d\n", rpacket->channel);
	printf("reliable: %d\n", rpacket->reliable);
	printf("length:   %d\n", rpacket->length);
	printf("data:     ");
	for(i = 0; i < rpacket->length; i++)
		printf("%02x ", rpacket->payload[i]);
	printf("\n");
#endif /* DEBUG */

	if(rpacket->payload[0] == 0xff) {
		msg = (struct csr_msg *)&(rpacket->payload[2]);
		cmd = (struct csr_bccmd *)msg->data;
#ifdef DEBUG
		printf("==CSR-Message==\n");
		printf("channel: %lu\n", CSR_GET_CH_ID(msg));
#endif /* DEBUG */
		if(CSR_GET_CH_ID(msg) == CSR_CH_ID_BCCMD)
			process_csr_event(cmd);
	}
}

/*
 * build a request packet to read one of the firmware revision fields
 */
void get_firmware_rev_info(struct ubcsp_packet *send_packet, uint16 id)
{
	struct cmd_pkt *pkt;
	struct csr_msg *msg;
	struct csr_bccmd *cmd;

	/* BCSP transport parameters */
	send_packet->channel = 5;
	send_packet->reliable = 1;
	send_packet->length = 1 + 5 * sizeof(uint16) + 6 * sizeof(uint16) +
		CMD_HDR_LEN;

	/* CSR specific header */
	pkt = (struct cmd_pkt *)(send_packet->payload);
	pkt->opcode = hci_put_opcode(0x00, MANUFACTURER_SPEC);
	pkt->len = 1 + 5 * sizeof(uint16) + 6 * sizeof(uint16);

	msg = (struct csr_msg *)pkt->data;
	cmd = (struct csr_bccmd *)msg->data;

	/* general msg header */
	CSR_SET_LAST(msg, 1);
	CSR_SET_FIRST(msg, 1);
	CSR_SET_CH_ID(msg, CSR_CH_ID_BCCMD);

	/* BCCMD type */
	cmd->type = CSR_MSGTYPE_GETREQ;
	cmd->len = 5 + 6;
	cmd->seq = csr_seq++;
	cmd->id = id;
	cmd->status = CSR_STATUS_OK;
	memset(cmd->data, 0x00, 6 * sizeof(uint16));
}

/*
 * build a request packet to ask for the size of a ps key entry
 */
void csr_pssize(struct ubcsp_packet *send_packet, uint16 pskey)
{
	struct cmd_pkt *pkt;
	struct csr_msg *msg;
	struct csr_bccmd *cmd;
	struct csr_bccmd_ps *ps;

	/* BCSP transport parameters */
	send_packet->channel = 5;
	send_packet->reliable = 1;
	send_packet->length = 1 + 5 * sizeof(uint16) + 4 * sizeof(uint16) +
		CMD_HDR_LEN;

	/* CSR specific header */
	pkt = (struct cmd_pkt *)(send_packet->payload);
	pkt->opcode = hci_put_opcode(0x00, MANUFACTURER_SPEC);
	pkt->len = 1 + 5 * sizeof(uint16) + 4 * sizeof(uint16);

	msg = (struct csr_msg *)pkt->data;
	cmd = (struct csr_bccmd *)msg->data;
	ps = (struct csr_bccmd_ps *)cmd->data;

	/* general msg header */
	CSR_SET_LAST(msg, 1);
	CSR_SET_FIRST(msg, 1);
	CSR_SET_CH_ID(msg, CSR_CH_ID_BCCMD);

	/* BCCMD type */
	cmd->type = CSR_MSGTYPE_GETREQ;
	cmd->len = 5 + 4;
	cmd->seq = csr_seq++;
	cmd->id = CSR_CMD_PS_SIZE;
	cmd->status = CSR_STATUS_OK;

	/* actual PS key request */
	ps->ps_key = pskey;
	ps->ps_len = 0;
	ps->unused = 0;
	ps->ps_val[0] = 0;
}

/*
 * construct request packet for reading one of the ps key entries
 */
void csr_getps(struct ubcsp_packet *send_packet, uint16 pskey, int size)
{
	struct cmd_pkt *pkt;
	struct csr_msg *msg;
	struct csr_bccmd *cmd;
	struct csr_bccmd_ps *ps;

	/* BCSP transport parameters */
	send_packet->channel = 5;
	send_packet->reliable = 1;
	send_packet->length = 1 + 5 * sizeof(uint16) + 3 * sizeof(uint16) +
		size * sizeof(uint16) + CMD_HDR_LEN;

	/* CSR specific header */
	pkt = (struct cmd_pkt *)(send_packet->payload);
	pkt->opcode = hci_put_opcode(0x00, MANUFACTURER_SPEC);
	pkt->len = 1 + 5 * sizeof(uint16) + 3 * sizeof(uint16) +
		size * sizeof(uint16);

	msg = (struct csr_msg *)pkt->data;
	cmd = (struct csr_bccmd *)msg->data;
	ps = (struct csr_bccmd_ps *)cmd->data;

	/* general msg header */
	CSR_SET_LAST(msg, 1);
	CSR_SET_FIRST(msg, 1);
	CSR_SET_CH_ID(msg, CSR_CH_ID_BCCMD);

	/* BCCMD type */
	cmd->type = CSR_MSGTYPE_GETREQ;
	cmd->len = 5 + 3 + size;
	cmd->seq = csr_seq++;
	cmd->id = CSR_CMD_PS;
	cmd->status = CSR_STATUS_OK;

	/* actual PS key request */
	ps->ps_key = pskey;
	ps->ps_len = size;
	ps->unused = 0;
	memset(ps->ps_val, 0x00, size * sizeof(uint16));
}

/*
 * set a ps key entry to a new value
 */
void csr_setps(struct ubcsp_packet *send_packet, uint16 pskey, int size,
	uint16 *data)
{
	struct cmd_pkt *pkt;
	struct csr_msg *msg;
	struct csr_bccmd *cmd;
	struct csr_bccmd_ps *ps;

	/* BCSP transport parameters */
	send_packet->channel = 5;
	send_packet->reliable = 1;
	send_packet->length = 1 + 5 * sizeof(uint16) + 3 * sizeof(uint16) +
		size * sizeof(uint16) + CMD_HDR_LEN;

	/* CSR specific header */
	pkt = (struct cmd_pkt *)(send_packet->payload);
	pkt->opcode = hci_put_opcode(0x00, MANUFACTURER_SPEC);
	pkt->len = 1 + 5 * sizeof(uint16) + 3 * sizeof(uint16) +
		size * sizeof(uint16);

	msg = (struct csr_msg *)pkt->data;
	cmd = (struct csr_bccmd *)msg->data;
	ps = (struct csr_bccmd_ps *)cmd->data;

	/* general msg header */
	CSR_SET_LAST(msg, 1);
	CSR_SET_FIRST(msg, 1);
	CSR_SET_CH_ID(msg, CSR_CH_ID_BCCMD);

	/* BCCMD type */
	cmd->type = CSR_MSGTYPE_SETREQ;
	cmd->len = 5 + 3 + size;
	cmd->seq = csr_seq++;
	cmd->id = CSR_CMD_PS;
	cmd->status = CSR_STATUS_OK;

	/* actual PS key request */
	ps->ps_key = pskey;
	ps->ps_len = size;
	ps->unused = 0;
	memcpy(ps->ps_val, data, size * sizeof(uint16));
}

/*
 * warm reboot the CSR module
 */
void csr_reboot(struct ubcsp_packet *send_packet)
{
	struct cmd_pkt *pkt;
	struct csr_msg *msg;
	struct csr_bccmd *cmd;

	/* BCSP transport parameters */
	send_packet->channel = 5;
	send_packet->reliable = 1;
	send_packet->length = 1 + 5 * sizeof(uint16) + 4 * sizeof(uint16) +
		CMD_HDR_LEN;

	/* CSR specific header */
	pkt = (struct cmd_pkt *)(send_packet->payload);
	pkt->opcode = hci_put_opcode(0x00, MANUFACTURER_SPEC);
	pkt->len = 1 + 5 * sizeof(uint16) + 4 * sizeof(uint16);

	msg = (struct csr_msg *)pkt->data;
	cmd = (struct csr_bccmd *)msg->data;

	/* general msg header */
	CSR_SET_LAST(msg, 1);
	CSR_SET_FIRST(msg, 1);
	CSR_SET_CH_ID(msg, CSR_CH_ID_BCCMD);

	/* BCCMD type */
	cmd->type = CSR_MSGTYPE_SETREQ;
	cmd->len = 5 + 4;
	cmd->seq = csr_seq++;
	cmd->id = CSR_WARM_RESET;
	cmd->status = CSR_STATUS_OK;
	memset(cmd->data, 0x00, 4 * sizeof(uint16));
}

/******************************************************************************/


int get_bdaddr(char* afilename, char* bdaddr)
{
	int n;
	const unsigned char tomtom_bdid[3] = {0x00, 0x13, 0x6c};
	unsigned char buffer[6];
	FILE*f=fopen(afilename,"rb");
	if (!f)
	{
		printf("Warning: could not open file with MAC address %s\n",afilename);
		return 0;  
	}
	n = fread(buffer,1,6,f);
	fclose(f);
	if (n != 6)
	{
		printf("Warning: MAC address was only %d bytes\n",n);
		goto FAIL_MAC_ADDRESS;
	}
	if (buffer[0] == 0xff)
	{
		printf("Warning: MAC address seems to be 0xff\n");
		goto FAIL_MAC_ADDRESS;
	}
	if (memcmp(buffer,tomtom_bdid,3) != 0) {
		printf("Warning: MAC address [0..2] incorrect 0x%02x 0x%02x 0x%02x\n",buffer[0],buffer[1],buffer[2]);
		if ((buffer[0] == 0x13) && (buffer[1] == 0x00)) {
			printf("Suggestion: swap 16-bit wise, buf[0] and buf[1], buf[2] and buf[3], buf[4] and buf[5]\n");
		}
		if ((buffer[5] == tomtom_bdid[0]) && (buffer[4] == tomtom_bdid[1]) && (buffer[3] == tomtom_bdid[2])) {
			printf("Suggestion: write the bytes from 0 to 5 instead of 5 to 0\n");
		}

		goto FAIL_MAC_ADDRESS;
	}
	memcpy(bdaddr,buffer,6);
	return 6;
FAIL_MAC_ADDRESS:
	printf("FAILED to read MAC addr succesfully from %s. Read = ",afilename);
	for (n=0; n<6; n++)
		printf("0x%02x ",buffer[n]);
	printf("\n");
	return 0;
}


/*
 * sets the BDADDR - uses the TomTom OUI 00-13-6c
 */
void set_bdaddr(struct ubcsp_packet *send_packet)
{
	uint8 bdaddr[] = {0x00, 0x13, 0x6c, 0x00, 0x00, 0x00};
	uint16 data[4];

	int fd;

	/* Read mac address from /mnt/flash/sysfile/mac */
	if(!get_bdaddr("/mnt/flash/sysfile/mac", bdaddr)) {
		if(!get_bdaddr("/mnt/flash/mactemp", bdaddr)) {
			printf("Could not open temporary mac file, creating "
				"one instead\n");

			fd = open("/dev/urandom", O_RDONLY);
			read(fd, &bdaddr[3], 3);
			close(fd);
			bdaddr[0] = 0x00;
			bdaddr[1] = 0x13;
			bdaddr[2] = 0x6c;

			FILE *f = fopen("/mnt/flash/mactemp", "wb");
			if(f) {
				fwrite(bdaddr, 1, 6, f);
				fclose(f);
			}
		} else {
			printf("Warning: Got mac address from temp mac file "
				"(some random number)\n");
		}
	}

	printf("Final BDADDR=%02X:%02X:%02X:%02X:%02X:%02X\n",
		bdaddr[0], bdaddr[1], bdaddr[2],
		bdaddr[3], bdaddr[4], bdaddr[5]);

	data[0] = bdaddr[3];
	data[1] = bdaddr[4] << 8 | bdaddr[5];
	data[2] = bdaddr[2];
	data[3] = bdaddr[0] << 8 | bdaddr[1];
	printf("Setting CSR BDADDR\n");
	csr_setps(send_packet, CSR_PS_BDADDR, 4, data);
}


/*
 * basic state machine
 * we start with state 10 and after every reply we received we continue with
 * the next one...
 * punchcard like state machine...
 */
void send_next_packet(struct ubcsp_packet *send_packet)
{
	uint16 data[76];
	int fd,i,num_ents;
	char *p;

	switch(state) {
		/**************************************************************/
		case 10:
			/* read chip family */
			printf("reading chip family\n");
			get_firmware_rev_info(send_packet, CSR_CMD_CHIPVER);
			state = 11;
			break;
		case 11:
			/* read firmware build id */
			printf("reading firmware build id\n");
			get_firmware_rev_info(send_packet, CSR_CMD_BUILD_ID);
			state = 12;
			break;
		case 12:
			/* read chip revision */
			printf("reading chip revision\n");
			get_firmware_rev_info(send_packet, CSR_CMD_CHIPREV);
			state = 13;
			break;
		case 13:
			// making sure sco uses hci in stead of pcm 
			memset(data, 0x00, sizeof(data));
			data[0]  = 0x0000;
			csr_setps(send_packet, CSR_PS_HOSTIO_MAP_SCO_PCM, 1, data);
			state = 14;
			break;
		case 14:
			// making sure cpu clock will not in to deep a power down mode during audio handling 
			memset(data, 0x00, sizeof(data));
			data[0]  = 0x0000;
			csr_setps(send_packet, 0x024D, 1, data);
			printf("min_cpu_clock %d\n",data[0]);
			state = 20;
			break;
		/**************************************************************/
		case 20:
			/* set external crystal frequency */
			memset(data, 0x00, sizeof(data));
			data[0] = (crystal_freq / 1000); /* ~16 or solid 26 MHz */
			csr_setps(send_packet, CSR_PS_ANA_FREQ, 1, data);
			state = 21;
			break;
		case 21:
			/* set bluetooth class of device */
			memset(data, 0x00, sizeof(data));
			data[0] = 0;
			csr_setps(send_packet, CSR_PS_CLASSOFDEVICE, 1, data);
			state = 22;
			break;
		case 22:
			/* set bluetooth module country code */
			memset(data, 0x00, sizeof(data));
			data[0] = 0x00; /* North America & Europe */
			csr_setps(send_packet, CSR_PS_COUNTRYCODE, 1, data);
			state = 23;
			break;
		case 23:
			/* configure CSR host interface protocol */
			switch (mode) {
				case BCCMDMODE_H4: 		
						p = "H4"; 		data[0] = 0x03; 
						break;
				case BCCMDMODE_UBCSP: 
						p = "UBCSP";	data[0] = 0x01;
						break;
				case BCCMDMODE_USB: 	
						p = "USB"; 		data[0] = 0x02;
						break;
				default:
						p = "";
						break;
			}

			printf("configuring host interface protocol to %s\n", p);
			csr_setps(send_packet, CSR_PS_HOST_INTERFACE, 1, data);
			state = 24;
			break;
		case 24:
			/* set default transmit power in dBm */
			memset(data, 0x00, sizeof(data));
			data[0] = 20; /* 20 dBm */
			csr_setps(send_packet, CSR_PS_LC_DEFAULT_TX_POWER,
				1, data);
			if (mode == BCCMDMODE_USB) {
				state = 40; /* skip baudrate stuff */
			} else {
				state = 25;
			}
			break;
		case 25:
			/* set baud rate */
			memset(data, 0x00, sizeof(data));
			printf("setting UART to %d baud (div %hu, want %d)\n",
				csr_rate, csr_div, real_baudrate);
			data[0] = csr_div;
			csr_setps(send_packet, CSR_PS_UART_BAUD_RATE,
				1, data);
			state = 26;
			break;
		case 26:
			/* disable RTS/CTS flow control */
			memset(data, 0x00, sizeof(data));
			data[0] = 0; /* disable RTS/CTS */
			csr_setps(send_packet, CSR_PS_UART_CONFIG_FLOW_CTRL_EN,
				1, data);
			state = 27;
			break;

		/******************************************************/	
		case 40:
			/* USB productid */
			memset(data, 0x00, sizeof(data));
			data[0] = 0x4d4a;
			csr_setps(send_packet, CSR_PS_USB_PRODUCTID, 1, data);
			state =  41;
			break;

		case 41:
			/* USB vendorid */
			memset(data, 0x00, sizeof(data));
			data[0] = 0x1390; /* TomTom USB vendorID */
			csr_setps(send_packet, CSR_PS_USB_VENDORID, 1, data);
			state =  27;
			break;

		case 27:
			/* setup external buffer capacitor on AIO(2) */
			// printf("enabling external filter capacitor\n");
			memset(data, 0x00, sizeof(data));
			data[0] = 1; /* read datasheet to understand... */
			csr_setps(send_packet, CSR_PS_AMUX_B, 1, data);
			state = 28;
			break;
		case 28:
			/* setup external PA control to 'non existing' */
			// printf("disabling external PA control\n");
			memset(data, 0x00, sizeof(data));
			data[0] = 0;
			csr_setps(send_packet, CSR_PS_TXRX_PIO_CONTROL,
				1, data);
			state = 29;
			break;
		case 29:
			/* setup the power table, see AN051 */
			memset(data, 0x00, sizeof(data));

			/* fill btcal_flash with 0xffff in case file can't be read */
			memset(&btcal_flash, 0xff, sizeof(btcal_flash));

			printf("generating CSR_PS_LC_POWER_TABLE\n");

			btcal_flash.power_table[0].u.external_pa = 0;
			btcal_flash.power_table[0].u.internal_pa = 14;
			btcal_flash.power_table[0].u.tx_dbm 	 = (uint8)-20;
			btcal_flash.power_table[0].u.reserved    = 0;

			btcal_flash.power_table[1].u.external_pa = 0;
			btcal_flash.power_table[1].u.internal_pa = 19;
			btcal_flash.power_table[1].u.tx_dbm    	 = (uint8)-16;
			btcal_flash.power_table[1].u.reserved    = 0;

			btcal_flash.power_table[2].u.external_pa = 0;
			btcal_flash.power_table[2].u.internal_pa = 24;
			btcal_flash.power_table[2].u.tx_dbm      = (uint8)-12;
			btcal_flash.power_table[2].u.reserved    = 0;

			btcal_flash.power_table[3].u.external_pa = 0;
			btcal_flash.power_table[3].u.internal_pa = 31;
			btcal_flash.power_table[3].u.tx_dbm      = (uint8)-8;
			btcal_flash.power_table[3].u.reserved    = 0;

			btcal_flash.power_table[4].u.external_pa = 0;
			btcal_flash.power_table[4].u.internal_pa = 39;
			btcal_flash.power_table[4].u.tx_dbm      = (uint8)-4;
			btcal_flash.power_table[4].u.reserved    = 0;

			btcal_flash.power_table[5].u.external_pa = 0;
			btcal_flash.power_table[5].u.internal_pa = 47;
			btcal_flash.power_table[5].u.tx_dbm      = (uint8)-0;
			btcal_flash.power_table[5].u.reserved    = 0;

			btcal_flash.power_table[6].u.external_pa = 0;
			btcal_flash.power_table[6].u.internal_pa = 54;
			btcal_flash.power_table[6].u.tx_dbm      = (uint8)3;
			btcal_flash.power_table[6].u.reserved    = 0;

			btcal_flash.power_table[7].u.external_pa = 0;
			btcal_flash.power_table[7].u.internal_pa = 63;
			btcal_flash.power_table[7].u.tx_dbm      = (uint8)6;
			btcal_flash.power_table[7].u.reserved    = 0;

			btcal_flash.power_table[8].raw = 0x00000000;

			num_ents = 0;
			for (i=0; i<9; i++) {
				data[i*2+0] = (btcal_flash.power_table[i].u.internal_pa << 8) |
							  btcal_flash.power_table[i].u.external_pa;
				data[i*2+1] = ( ((uint8)btcal_flash.power_table[i].u.tx_dbm) << 8)|
							  btcal_flash.power_table[i].u.reserved;
				if (btcal_flash.power_table[i].raw != 0x00000000)
					num_ents++;
				else 
					break; // last entry
			}
			if (btcal_sanity_check() == 0) {
#ifdef SEND_POWERTABLE
				printf("Sending powertable, %d entries\n", num_ents);
				csr_setps(send_packet, CSR_PS_LC_POWER_TABLE, num_ents * 2, &data[0]); 
#else
				printf("Skipping powertable\n");
#endif
			} else {
				printf("NOT sending BT powertable: flash powertable sanity check failed.\n");
			}
			state = 30;
			break;
		/**************************************************************/
		case 30:
			/* get storage size of BDADDR ps key */
			csr_pssize(send_packet, CSR_PS_BDADDR);
			state = 31;
			break;
		case 31:
			/* read BDADDR */
			csr_getps(send_packet, CSR_PS_BDADDR, pskey_size);
			state = 32;
			break;
		case 32:
			/* set BDADDR to TomTom OUI and serial number */
			set_bdaddr(send_packet);
			state = 33;
			break;
		case 33:
			/* read ana_ftrim from /mnt/flash/sysfile/btcal */
			/* this is a file mapped to flash offset 0x0000014C */
			/* format of this offset data is:     
				2 bytes Crystal trim 
				4 bytes Power table 1 (deprecated)
				4 bytes Power table 2 (deprecated)
				406 bytes up to power table 60
			*/
			if (ext_clock) {
				printf("Setting ANA_FTRIM to 0, external clock\n");
				btcal_flash.crystal_trim = 0;
			} else {
				fd = open("/mnt/flash/sysfile/btcal", O_RDONLY);
				if (fd == -1) {
					printf("Err opening btcal file\n");
					btcal_flash.crystal_trim = 0xffff;
				} else {
					/* 2 bytes ANA_FTRIM, +2 deprecated pt entries */
					const int pt_flash_len = 2;
					if (read(fd, &btcal_flash, pt_flash_len) != pt_flash_len) {
						printf("Err reading btcal file %d bytes\n", pt_flash_len);
						btcal_flash.crystal_trim = 0xffff;
					} else {
						// Only read ANA_FTRIM
						// rest (8 bytes) is deprecated powertable data
						printf("Read ANA_FTRIM %04x\n", btcal_flash.crystal_trim);
						btcal_flash.crystal_trim &= 0x00ff;
						printf("Truncated to ANA_FTRIM %04x\n", btcal_flash.crystal_trim);
					}
				}
			}
			/* set PS_ANA_FTRIM, 0-63, 1 bit = ~110 femtoFarad */
			/* was read from flash in state 21 */
			/* was read from flash in state 21 */
			data[0] = btcal_flash.crystal_trim;
			if (data[0] > 0x3f) {
				printf("Invalid btcal, centering ANA_FTRIM\n");
				printf("WARNING: bad bluetooth performance\n");
				data[0] = 31; /* ~center of 0-63 */
			}
			csr_setps(send_packet, CSR_PS_ANA_FTRIM, 1, data);
			printf("setting ANA_FTRIM %d * ~110 femtoF\n", data[0]);
			state = 34;
			break;

		/**************************************************************/
		case 34:
			/* set bluetooth features */
 			/* original configuration: 0xff 0xff 0x8f 0xf8 0x18 0x18 0x00 0x80
 			 * modified configuration: 0xff 0xff 0x8f 0x78 0x18 0x18 0x00 0x80
			 * this is chip dependent data, so needs to be derived from original feature configuration,
			 * original feature configuration can be obtained via HCI interface using following sequence:
			 * 1) disable this case statement (should go to the next state immediatelly)
			 * 2) recompile and reinstalling csrinit on the GO
			 * 3) running rc.suspend and then rc.resume from /etc to update csr chip configuration
			 * 4) run "hciconfig hci0 features" which will print original hex feature configuration string
			 * 5) modify the string to disable EHV3 support, bit 24 set to 0; same for EV5, EV4 ? bits 38,39? (eSCO)
			 * 6) update the data[] assignments below and recompile/reinstall csrinit */
			// printf("setting bluetooth features of device\n");0
			memset(data, 0x00, sizeof(data));
			data[0] = 0xffff;
			data[1] = 0x788f;
			data[2] = 0x1818;
			data[3] = 0x8000;
			csr_setps(send_packet, CSR_PS_FEATURES, 4, data);
			state = 35;
			break;
			
		case 35:
			/* increase SCO buffer */
			memset(data, 0x00, sizeof(data));
			data[0] = 0x000a;
			data[1] = 0x00a0;
			csr_setps(send_packet, CSR_PS_HOSTIO_SCO_HCI_THRESHOLDS, 2,
				data);
			state = 36;
			break;
		case 36:
			/* max number of open SCO channels is 2 */
			memset(data, 0x00, sizeof(data));
			data[0] = 0x0002;
			csr_setps(send_packet, CSR_PS_MAX_SCOS, 1,
				data);
			if (detected_chiprevision == CHIPREVISION_BLUECORE3_ROM) {
				if ((csr_build_id == 1298) || (csr_build_id == 1593))
					state = 55; /* patch BC3ROM */
			} else if (detected_chiprevision == CHIPREVISION_BLUECORE4_ROM) {
				//if ((csr_build_id == 1298) || (csr_build_id == 1593))
					state = 60; /* patch BC4ROM V. */
			} else {
				/* just reboot */
				state = 97;
			}
			break;
		case 55:
			/* patch BC3ROM, part 1 */
			printf("Patching BC3ROM firmware (patch 22, buildid %d)\n", csr_build_id);
			
			memcpy(&data[0], patch_22, sizeof(patch_22));
			if (csr_build_id == 1298) data[36] = 0xec80;
			if (csr_build_id == 1593) data[36] = 0xeda7;
			csr_setps(send_packet, CSR_PS_PATCH_22, 36, data);

			state = 60; // just reboot
			break;
			
		/**************************************************************/
		// patch cores when needed
		case 60:
// patch_spare5, PSKEY_PATCH34 (0x00b8, 184)
data[0] = 0x0002; 
data[1] = 0x8416;
data[2] = 0x031b;
data[3] = 0x4216;
data[4] = 0x22f4;
data[5] = 0x4316;
data[6] = 0x0684;
data[7] = 0x1ff0;
data[8] = 0x9015;
data[9] = 0x08c4;
data[10] = 0x16f0;
data[11] = 0x9511;
data[12] = 0x9515;
data[13] = 0xfdc4;
data[14] = 0x9525;
data[15] = 0x9015;
data[16] = 0x08b4;
data[17] = 0x9025;
data[18] = 0x9715;
data[19] = 0x01b4;
data[20] = 0x9725;
data[21] = 0xff00;
data[22] = 0x6f15;
data[23] = 0xfdc4;
data[24] = 0xff00;
data[25] = 0x6f25;
data[26] = 0xff00;
data[27] = 0x5315;
data[28] = 0x20b4;
data[29] = 0xff00;
data[30] = 0x5325;
data[31] = 0x9521;
data[32] = 0x0118;
data[33] = 0xff2b;
data[34] = 0xff0e;
data[35] = 0x8400;
data[36] = 0x1218;
data[37] = 0x00e2;
data[38] = 0x0113;
data[39] = 0x3f30;
data[40] = 0xe721;
data[41] = 0x0118;
data[42] = 0xff2b;
data[43] = 0xff0e;
data[44] = 0x8400;
data[45] = 0x1918;
data[46] = 0x00e2;
data[47] = 0x5d46;
			csr_setps(send_packet, 0x00b8, 48, data);
			
			state = 61;
			break;

		case 61:
// patch_spare4, PSKEY_PATCH33 (0x00b7, 183)
data[0] = 0x0003;
data[1] = 0x3041;
data[2] = 0x0014;
data[3] = 0xfe27;
data[4] = 0x0114;
data[5] = 0xff27;
data[6] = 0x0218;
data[7] = 0xff2b;
data[8] = 0xff0e;
data[9] = 0x3000;
data[10] = 0x4418;
data[11] = 0x00e2;
data[12] = 0x63c2;
			csr_setps(send_packet, 0x00b7, 13, data);

			state = 62;
			break;

		case 62:
// patch_spare6, PSKEY_PATCH35 (0x00b9, 185)
data[0] = 0x0002;
data[1] = 0xb6bf;
data[2] = 0x0117;
data[3] = 0xb800;
data[4] = 0xe184;
data[5] = 0x07f4;
data[6] = 0xb800;
data[7] = 0xf184;
data[8] = 0x04f4;
data[9] = 0xb800;
data[10] = 0x1784;
data[11] = 0x12f0;
data[12] = 0xe300;
data[13] = 0xb519;
data[14] = 0x06f4;
data[15] = 0x1412;
data[16] = 0x15b2;
data[17] = 0x0cf0;
data[18] = 0x001a;
data[19] = 0xfcf0;
data[20] = 0xfc3c;
data[21] = 0x0014;
data[22] = 0x0327;
data[23] = 0x0118;
data[24] = 0xff2b;
data[25] = 0xff0e;
data[26] = 0x4400;
data[27] = 0x7818;
data[28] = 0x00e2;
data[29] = 0x043c;
data[30] = 0xfe0e;
data[31] = 0xffe3;
data[32] = 0x8cd3;
			csr_setps(send_packet, 0x00b9, 33, data);

			state = 63;
			break;

		case 63:
// patch_bc_boot, PSKEY_PATCH1 (0x0097, 151)
data[0] = 0x045c;
data[1] = 0x032b;
data[2] = 0x020a;
data[3] = 0x0014;
data[4] = 0x0027;
data[5] = 0x2100;
data[6] = 0xb434;
data[7] = 0x0018;
data[8] = 0x012b;
data[9] = 0x010e;
data[10] = 0xf100;
data[11] = 0x4018;
data[12] = 0x009e;
data[13] = 0x0084;
data[14] = 0x10f4;
data[15] = 0xf225;
data[16] = 0xf219;
data[17] = 0x0116;
data[18] = 0x0012;
data[19] = 0x001b;
data[20] = 0x003b;
data[21] = 0x003b;
data[22] = 0xfd00;
data[23] = 0x7a38;
data[24] = 0x0022;
data[25] = 0x0126;
data[26] = 0xf215;
data[27] = 0xa000;
data[28] = 0xfe54;
data[29] = 0x0226;
data[30] = 0x0017;
data[31] = 0x0134;
data[32] = 0x0027;
data[33] = 0x0884;
data[34] = 0xe3fc;
data[35] = 0x2100;
data[36] = 0xbc14;
data[37] = 0x0018;
data[38] = 0x012b;
data[39] = 0x010e;
data[40] = 0xf100;
data[41] = 0x4018;
data[42] = 0x009e;
data[43] = 0x0084;
data[44] = 0x08f4;
data[45] = 0x1410;
data[46] = 0x0123;
data[47] = 0x010e;
data[48] = 0xa000;
data[49] = 0x0054;
data[50] = 0x0127;
data[51] = 0x019f;
data[52] = 0x020e;
data[53] = 0x043c;
data[54] = 0xffe3;
data[55] = 0x2a15;
			csr_setps(send_packet, 0x0097, 56, data);

			state = 64;
			break;

		case 64:
// patch_spare7, PSKEY_PATCH36 (0x00ba, 186)
data[0] = 0x0001;
data[1] = 0x03f3;
data[2] = 0x1414;
data[3] = 0xde00;
data[4] = 0xbc25;
data[5] = 0xe000;
data[6] = 0x6b15;
data[7] = 0xa000;
data[8] = 0xe354;
data[9] = 0xde00;
data[10] = 0xbd25;
data[11] = 0x0114;
data[12] = 0xfe00;
data[13] = 0x8c25;
data[14] = 0x0400;
data[15] = 0x3e14;
data[16] = 0xfe00;
data[17] = 0x8d25;
data[18] = 0xfe00;
data[19] = 0x8e15;
data[20] = 0x3134;
data[21] = 0xfe00;
data[22] = 0x8e25;
data[23] = 0x0018;
data[24] = 0xff2b;
data[25] = 0xff0e;
data[26] = 0x0400;
data[27] = 0xf918;
data[28] = 0x00e2;
data[29] = 0xff2b;
data[30] = 0x0084;
data[31] = 0x11ec;
data[32] = 0x9811;
data[33] = 0x4000;
data[34] = 0x00c0;
data[35] = 0xe921;
data[36] = 0x0010;
data[37] = 0x9025;
data[38] = 0x9815;
data[39] = 0x4000;
data[40] = 0x00c4;
data[41] = 0xe9d5;
data[42] = 0x05f0;
data[43] = 0x0130;
data[44] = 0x4000;
data[45] = 0x0180;
data[46] = 0xf8fc;
data[47] = 0xffe3;
data[48] = 0x0008;
data[49] = 0x0000;
data[50] = 0xffe3;
data[51] = 0x0114;
data[52] = 0xfe00;
data[53] = 0x8c25;
data[54] = 0x0400;
data[55] = 0xf314;
data[56] = 0xfe00;
data[57] = 0x8d25;
data[58] = 0xfe00;
data[59] = 0x8e15;
data[60] = 0x3154;
data[61] = 0xfe00;
data[62] = 0x8e25;
data[63] = 0xde00;
data[64] = 0xbd19;
data[65] = 0xa000;
data[66] = 0x1338;
data[67] = 0x0814;
data[68] = 0x0026;
data[69] = 0xffe3;
data[70] = 0x389c;
			csr_setps(send_packet, 0x00ba, 71, data);

			state = 65;
			break;

		case 65:
// patch_spare3, PSKEY_PATCH32 (0x00b6, 182)
data[0] = 0x0001;
data[1] = 0x23bf;
data[2] = 0xe700;
data[3] = 0xb015;
data[4] = 0x04f0;
data[5] = 0xff00;
data[6] = 0xffc0;
data[7] = 0x03e0;
data[8] = 0x0100;
data[9] = 0x00b0;
data[10] = 0xe721;
data[11] = 0xe515;
data[12] = 0x0018;
data[13] = 0xff2b;
data[14] = 0xff0e;
data[15] = 0x2400;
data[16] = 0xd718;
data[17] = 0x00e2;
data[18] = 0x7d30;
			csr_setps(send_packet, 0x00b6, 19, data);

			state = 67;
			break;

// patch_38, PSKEY_PATCH38 (0x00bc, 188)
// &00bc = 9511 fdc0 9515 9521 9711 4000 00b0 9721 6111 04b0 6121 ff00 5211 01b0 ff00 5221 9525 00e2 db18
			data[0]  = 0x0000;
			csr_setps(send_packet, 0x00bc, 1, data);

		case 67:
// patch_spare2, PSKEY_PATCH31 (0x00b5, 181)
data[0] = 0x0001;
data[1] = 0xf4c6;
data[2] = 0xfe0a;
data[3] = 0xff2b;
data[4] = 0xfb3c;
data[5] = 0x3784;
data[6] = 0x03e4;
data[7] = 0x0010;
data[8] = 0x02e0;
data[9] = 0x0110;
data[10] = 0xe700;
data[11] = 0xb021;
data[12] = 0x0018;
data[13] = 0xff2b;
data[14] = 0xff0e;
data[15] = 0xf500;
data[16] = 0xc918;
data[17] = 0x00e2;
data[18] = 0x86b9;
			csr_setps(send_packet, 0x00b5, 19, data);

			state = 68;
			break;

		case 68:
// patch_37, PSKEY_PATCH37 (0x00bb, 187)
data[0] = 0x0002;
data[1] = 0x0ec1;
data[2] = 0x0018;
data[3] = 0xff2b;
data[4] = 0xff0e;
data[5] = 0x6800;
data[6] = 0xa718;
data[7] = 0x009e;
data[8] = 0x023c;
data[9] = 0xfe0e;
data[10] = 0xffe3;
data[11] = 0xead4;
			csr_setps(send_packet, 0x00bb, 12, data);

			state = 69;
			break;

		case 69:
// patch_radiotest_tx_start, PSKEY_PATCH17 (0x00a7, 167)
data[0] = 0x0513;
data[1] = 0x00f6;
data[2] = 0xff00;
data[3] = 0x5311;
data[4] = 0xe700;
data[5] = 0xb015;
data[6] = 0x04f0;
data[7] = 0xff00;
data[8] = 0xffc0;
data[9] = 0x03e0;
data[10] = 0x0100;
data[11] = 0x00b0;
data[12] = 0xff00;
data[13] = 0x5321;
data[14] = 0x00e2;
data[15] = 0xbc45;
			csr_setps(send_packet, 0x00a7, 16, data);

			state = 70;
			break;

		case 70:
// patch_lc_rx_hw_set, PSKEY_PATCH25 (0x00af, 175)
data[0] = 0x0915;
data[1] = 0x0184;
data[2] = 0x00f2;
data[3] = 0x0214;
data[4] = 0x0925;
data[5] = 0x00e2;
data[6] = 0xf019;
			csr_setps(send_packet, 0x00af, 7, data);

			state = 71;
			break;

		case 71:
// patch_spare1, PSKEY_PATCH30 (0x00b4, 180)
data[0] = 0x0002;
data[1] = 0xb264;
data[2] = 0x1414;
data[3] = 0x0118;
data[4] = 0xff2b;
data[5] = 0xff0e;
data[6] = 0xb200;
data[7] = 0x6718;
data[8] = 0x00e2;
data[9] = 0x27ff;
			csr_setps(send_packet, 0x00b4, 10, data);


			state = 97;
			break;

		case 97:
			/* enable PS key protection */
			memset(data, 0x00, sizeof(data));
			data[0] = 1;
			csr_setps(send_packet, CSR_PS_BCCMD_SECURITY_ACTIVE, 1,
				data);
			state = 98;
			break;
		case 98:
			/* warm reboot CSR chip */
			printf("warm reboot\n");
			csr_reboot(send_packet);
			state = 99;
			break;
		case 99:
			return;
	}

#ifdef DEBUG
	{
		int i;
		printf("\n--- sending packet: ---\n");
		printf("channel:  %d\n", send_packet->channel);
		printf("reliable: %d\n", send_packet->reliable);
		printf("length:   %d\n", send_packet->length);
		printf("data:     ");
		for(i = 0; i < send_packet->length; i++)
			printf("%02x ", send_packet->payload[i]);
		printf("\n");
	}
#endif /* DEBUG */
}

void do_pretty_amazing_usb_magic()
{
	struct ubcsp_packet send_packet;
	uint8 send_buffer[512];
	struct ubcsp_packet receive_packet;
	uint8 receive_buffer[512];

	memset(&send_packet, 0, sizeof send_packet);
	memset(&receive_packet, 0, sizeof receive_packet);

	/* setup the receive packet payload */
	send_packet.length = 512;
	send_packet.payload = send_buffer;

	/* setup the packets payload pointer */
	receive_packet.length = 512;
	receive_packet.payload = receive_buffer;

	// TODO: move PS statemachine stuff to other file
	state = 10; // 33 ;

	/* pick up a waiting HCI status packet, ignore */
	usb_receive_packet(&receive_packet);
	
	while (state != 99) {
		/* prepare next packet */
		send_next_packet(&send_packet);
		/* TODO: send USB HCI commands */
		send_usb_packet(&send_packet.payload[0], send_packet.length); /* offset 8 should be the 'desc' field of csr_msg */
	
		usb_receive_packet(&receive_packet);
		process_packet(&receive_packet);
	}
	
}

int detect_nop(struct ubcsp_packet *rpacket)
{
	static char nop_frame[6] = {0x0f, 0x04, 0x00, 0x01, 0x00, 0x00};
	return (memcmp(rpacket->payload, nop_frame, 6) == 0);
}


/*
 * do one complete run through all states
 */
void do_pretty_amazing_magic()
{
	char activity;
	int delay = 0;
	int timeout;

	struct ubcsp_packet send_packet;
	uint8 send_buffer[512];
	struct ubcsp_packet receive_packet;
	uint8 receive_buffer[512];

	memset(&send_packet, 0, sizeof send_packet);
	memset(&receive_packet, 0, sizeof receive_packet);

	/* setup the receive packet payload */
	send_packet.length = 512;
	send_packet.payload = send_buffer;

	/* setup the packets payload pointer */
	receive_packet.length = 512;
	receive_packet.payload = receive_buffer;

	state = 10;

	/* ask ubcsp to receive a packet -- initiates the link establishment */
	ubcsp_receive_packet(&receive_packet);

	/* start our punchcard... */
	timeout = 0;
	int ready = 0, pending_send = 0;
	while(1) {
		delay = ubcsp_poll(&activity);
		if (!ready && (activity & UBCSP_PACKET_SENT))
			pending_send = 1;
		if(activity & UBCSP_PACKET_RECEIVED) {
			/* process the newly arrived packet */
			process_packet(&receive_packet);
			if (detect_nop(&receive_packet)) {
				ready = 1;
				if (pending_send)
					activity |= UBCSP_PACKET_SENT;
			}
			/* reschedule the receive buffer */
			receive_packet.length = 512;
			ubcsp_receive_packet(&receive_packet);

			timeout = 0;
		}
		if(ready && (activity & UBCSP_PACKET_SENT)) {
			/* advance punchcard one step... */
			send_next_packet(&send_packet);
			ubcsp_send_packet(&send_packet);
			timeout = 0;
		}

		if(delay) {
			/* sleep some 100ms */
			usleep(10000);
			timeout++;
			if((timeout >= 1) && (state == 99)) {
				fprintf(stderr, "configuration complete\n");
				return;
			}
			if(timeout >= 1000) {
				fprintf(stderr, "timed out waiting for a packet\n");
				return;
			}
		}
	}
}

static inline int divrnd(int nom, int den)
{
	return (nom + den / 2) / den;
}

static int calc_real_rate(int rate)
{
	const char clocks[] = "/proc/clocks";
	int res = rate;
	FILE *fp;
	int pclk;
	char buf[128];
	int div;

	/* Read PCLK from /proc/clocks */
	if ((fp = fopen(clocks, "r")) == NULL) {
		warn("Could not open %s", clocks);
	} else {
		for (pclk = 0; !feof(fp) && !ferror(fp);) {
			if (fgets(buf, sizeof buf, fp) == NULL)
				break;
			if (sscanf(buf, "pclk: %d", &pclk) == 1)
				break;
		}
		if (ferror(fp)) {
			warn("Could not read %s", clocks);
		} else if (pclk == 0) {
			warnx("Could not find pclk in %s", clocks);
		} else if (pclk < 1000000 || pclk > 1000000000) {
			warnx("Insane pclk %d in %s", pclk, clocks);
		} else {
			/* The baud rate divisor is determined by: PCLK / (16 * rate) */
			div = divrnd(pclk, 16 * rate);
			/* Therefore, the real rate is PCLK / (16 * divisor) */
			res = divrnd(pclk, 16 * div);
			DBG("rate %d, pclk %d, div %d -> real rate %d\n",
				rate, pclk, div, res);
		}
		fclose(fp);
	}

	return res;
}

#define CSR_MULT				128		/*       65536 / 512 */
#define CSR_CLOCK_16000	31250	/*.00 16000000 / 512 */
#define CSR_CLOCK_26000	50718	/*.25 26000000 / 512 */

static inline uint16 calc_csr_div(int rate)
{
	/* only supports 16000 or 26000 Mhz now... */
	return divrnd(rate * CSR_MULT, (crystal_freq==26000000)?CSR_CLOCK_26000:CSR_CLOCK_16000);
}

static inline int calc_csr_rate(uint16 div)
{
	switch (crystal_freq) {
		default:
		case 16000000:
			return divrnd((int) div * CSR_CLOCK_16000, CSR_MULT);
		case 26000000:
			return divrnd((int) div * CSR_CLOCK_26000, CSR_MULT);
	}
}

/******************************************************************************/


#define GPIO_DRIVER_MAGIC				'U'
#define IOR_HWSTATUS					_IOR(GPIO_DRIVER_MAGIC, 0, HARDWARE_STATUS)
#define IOR_DISK_ACCESS       _IOR(GPIO_DRIVER_MAGIC, 7, unsigned int)
#define IOR_BT_MODE			      _IOR(GPIO_DRIVER_MAGIC, 8, unsigned int)
 
/* btmode: 1 == USB, 0 == BCSP */
int go_reset_bt(int btmode)
{
	int fd;
	const char *hwstatus = "/dev/hwstatus";
	/* reboot CSR chip in USB mode */
	fd = open(hwstatus, O_RDWR | O_NOCTTY);
	if (fd == -1) {
		printf("can't open %s\n", hwstatus);
		exit(-1);
	} else {
		if (ioctl(fd, IOR_BT_MODE, &btmode) == -1) {
			printf("can't ioctl %s\n", hwstatus);
		}
		close(fd);
		printf("BT held in %s mode\n", btmode ? "usb" : "bcsp");
		/* todo: wait until new USB device reports itself */
		usleep(1000000);
		printf("checking...\n");
	}
	return 0;
}

/******************************************************************************/

/*
 * main function
 */
int main(int argc, char *argv[])
{
	int ret;
	int baudrate;
	speed_t speed;

	csr_seq = 0;

	/* parse command line */
	if (argc != 5) {
usage:
		fprintf(stderr, "csrinit 3.0\n");
		fprintf(stderr, "Usage: %s bcsp|h4|usb devname speed extclkspeed\n", argv[0]);
		fprintf(stderr, "bcsp and h4 imply serial devname\n");
		return 1;
	}

	if (argc == 5) {
		if (strcmp(argv[1], "bcsp") == 0) {
			// normal serial mode
			mode = BCCMDMODE_UBCSP;
		} else if (strcmp(argv[1], "h4") == 0) {
			mode = BCCMDMODE_H4;
		} else if (strcmp(argv[1], "usb") == 0) {
			mode = BCCMDMODE_USB;
		} else
			goto usage;
	}

	if (mode != BCCMDMODE_USB) {
		go_reset_bt(0);
		baudrate = atoi(argv[3]);
		switch (baudrate) {
		case 50:      speed = B50;      break;
		case 75:      speed = B75;      break;
		case 110:     speed = B110;     break;
		case 134:     speed = B134;     break;
		case 150:     speed = B150;     break;
		case 200:     speed = B200;     break;
		case 300:     speed = B300;     break;
		case 600:     speed = B600;     break;
		case 1200:    speed = B1200;    break;
		case 1800:    speed = B1800;    break;
		case 2400:    speed = B2400;    break;
		case 4800:    speed = B4800;    break;
		case 9600:    speed = B9600;    break;
		case 19200:   speed = B19200;   break;
		case 38400:   speed = B38400;   break;
		case 57600:   speed = B57600;   break;
		case 115200:  speed = B115200;  break;
		case 230400:  speed = B230400;  break;
		case 460800:  speed = B460800;  break;
		case 500000:  speed = B500000;  break;
		case 576000:  speed = B576000;  break;
		case 921600:  speed = B921600;  break;
		case 1000000: speed = B1000000; break;
		case 1152000: speed = B1152000; break;
		case 1500000: speed = B1500000; break;
		case 2000000: speed = B2000000; break;
		case 2500000: speed = B2500000; break;
		case 3000000: speed = B3000000; break;
		case 3500000: speed = B3500000; break;
		case 4000000: speed = B4000000; break;
		default: errx(1, "Invalid baudrate %s", argv[3]);
		}
	
		fprintf(stderr,"going for baudrate %d\n", baudrate);
		real_baudrate = calc_real_rate(baudrate);
		fprintf(stderr,"got real baudrate %d\n", real_baudrate);
		csr_div = calc_csr_div(real_baudrate);
		csr_rate = calc_csr_rate(csr_div);
		fprintf(stderr,"got csr divisor %hu, baudrate %d\n", csr_div, csr_rate);

		/* open serial port */
		if ((serialfd = serial_open(argv[2], speed)) < 0)
			return 1;
	} else {
		/* HCI over USB mode */
		go_reset_bt(1);
	}

	switch (atoi(argv[4])) {
		case 16000000:
		case 26000000:
			/* good */
			crystal_freq = atoi(argv[4]);
			fprintf(stderr, "extclkspeed %d\n",atoi(argv[4]));
			if (crystal_freq != 16000000) {
				ext_clock = 1;
			}
			break;
		default:
			fprintf(stderr, "invalid extclkspeed %d\n",atoi(argv[4]));
			goto usage;
	}

	/* initialize BCSP stack */
	/* some state might also be needed for USB */
	ubcsp_initialize();
	ret = 1; /* error */

	/* configure the CSR BlueCore */
	if (mode != BCCMDMODE_USB) {
		/* do h4 of ubcsp over serial */
		do_pretty_amazing_magic();
		/* now, wait until bluez picks up USB device */
		return 0;
	} else {
		/* use libusb to turn the BCXROM into a real BT-HCI device */
		ret = open_usb_conn(); /* argv[2] ingored */
		switch (ret) {
			case 0:
				do_pretty_amazing_usb_magic();
				ret = 0;
				/* succesfull ? wait until device is alive */
				usleep(1000000);
				/* fallthrough */
			case 1:
			default:
				close_usb_conn();
				return ret;
		}
	}
}
